import itertools
from collections.abc import Sequence

import pytest

from musiclib.note import Note
from musiclib.note import SpecificNote
from musiclib.noteset import NoteSet
from musiclib.noteset import bits_to_intervals
from musiclib.noteset import intervals_to_bits
from musiclib.scale import Scale
from musiclib.scale import all_scales


def test_empty():
    NoteSet(frozenset())
    with pytest.raises(KeyError):
        NoteSet(frozenset(), root=Note('C'))


@pytest.mark.parametrize(
    'bits, intervals', (
        ('101011010101', frozenset({0, 2, 4, 5, 7, 9, 11})),
        ('110101101010', frozenset({0, 1, 3, 5, 6, 8, 10})),
        ('101001010100', frozenset({0, 2, 5, 7, 9})),
        ('101101010010', frozenset({0, 2, 3, 5, 7, 10})),
        ('000000000000', frozenset()),
    ),
)
def test_bits_intervals(bits, intervals):
    assert bits_to_intervals(bits) == intervals
    assert intervals_to_bits(intervals) == bits


@pytest.mark.parametrize(
    'string, expected', (
        ('CDEFGAB/C', NoteSet(frozenset(map(Note, 'CDEFGAB')), root=Note('C'))),
        ('CDEFGAB', NoteSet(frozenset(map(Note, 'CDEFGAB')))),
        ('CdeFGab/e', NoteSet(frozenset(map(Note, 'CdeFGab')), root=Note('e'))),
        ('CEG/C', NoteSet(frozenset(map(Note, 'CEG')), root=Note('C'))),
        ('fa/a', NoteSet(frozenset(map(Note, 'fa')), root=Note('a'))),
        ('', NoteSet(frozenset())),
    ),
)
def test_from_str(string, expected):
    assert NoteSet.from_str(string) == expected


@pytest.mark.parametrize(
    'noteset, intervals', (
        (NoteSet.from_str('CDEFGAB/C'), (0, 2, 4, 5, 7, 9, 11)),
        (NoteSet.from_str('DeFGAbC/D'), (0, 1, 3, 5, 7, 8, 10)),
        (NoteSet(frozenset()), ()),
    ),
)
def test_intervals(noteset, intervals):
    assert noteset.intervals_ascending == intervals
    assert noteset.intervals == frozenset(intervals)


@pytest.mark.parametrize(
    'noteset, note_to_interval', [
        (NoteSet.from_str('CDE'), {}),
        (NoteSet.from_str('CDE/C'), {Note('C'): 0, Note('D'): 2, Note('E'): 4}),
    ],
)
def test_note_to_interval(noteset, note_to_interval):
    assert noteset.note_to_interval == note_to_interval


@pytest.mark.parametrize(
    'notes, bits', (
        ('CDEFGAB/C', '101011010101'),
        ('dfb/d', '100001000100'),
        ('', '000000000000'),
    ),
)
def test_bits(notes, bits):
    assert NoteSet.from_str(notes).bits == bits


@pytest.mark.parametrize(
    'noteset, bits', (
        (NoteSet.from_str('CDE'), (1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0)),
        (NoteSet.from_str('df'), (0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0)),
    ),
)
def test_bits_notes(noteset, bits):
    assert noteset.bits_notes == bits


@pytest.mark.parametrize(
    'value', (
        'CDE',
        set(map(Note, 'CDE')),
        tuple(map(Note, 'CDE')),
        list(map(Note, 'CDE')),
    ),
)
def test_notes_type_is_frozenset(value):
    with pytest.raises(TypeError):
        NoteSet(value)


def test_contains():
    assert Note('C') in NoteSet.from_str('C')
    assert Note('C') not in NoteSet.from_str('D')
    assert NoteSet.from_str('CD') <= NoteSet.from_str('CDE')
    assert NoteSet.from_str('CDE') <= NoteSet.from_str('CDE')
    assert not NoteSet.from_str('CDEF') <= NoteSet.from_str('CDE')
    empty_noteset = NoteSet(frozenset())
    assert Note('C') not in empty_noteset
    assert empty_noteset <= NoteSet.from_str('CDE')


def test_root_validation():
    with pytest.raises(KeyError):
        NoteSet.from_str('AB/E')


def test_note_i():
    fs = frozenset(map(Note, 'CDEfGaB'))
    noteset = NoteSet(fs)
    assert fs == noteset.note_i.keys()
    assert noteset.note_i[Note('C')] == 0
    assert noteset.note_i[Note('B')] == 6
    assert noteset.note_i[Note('f')] == 3
    assert noteset.note_i[Note('G')] == 4
    assert NoteSet(frozenset()).note_i == {}


@pytest.mark.parametrize(
    'intervals, root, expected', (
        (frozenset({0, 4, 7}), 'C', NoteSet.from_str('CEG/C')),
        (frozenset({0, 1, 3, 5, 7, 8, 10}), 'E', NoteSet.from_str('CDEFGAB/E')),
        (frozenset({0, 2, 3, 5, 7, 9, 10}), 'f', NoteSet.from_str('faABdeE/f')),
        (frozenset(), None, NoteSet(frozenset(), root=None)),
    ),
)
def test_from_intervals(intervals, root, expected):
    assert NoteSet.from_intervals(intervals, root) is expected


def test_subclasses_names_unreachable():
    with pytest.raises(KeyError):
        NoteSet.from_name('C', 'major')  # test that Scale names are unreachable
    with pytest.raises(KeyError):
        NoteSet.from_name('e', 'aug')  # test that Chord names are unreachable


@pytest.mark.parametrize(
    'noteset, notes_octave_fit', (
        (NoteSet.from_str('efGd'), 'defG'),
        (NoteSet.from_str('efGd/e'), 'defG'),
        (NoteSet.from_str('FGbBCd/F'), 'CdFGbB'),
        (NoteSet(frozenset()), ''),
    ),
)
def test_notes_octave_fit(noteset, notes_octave_fit):
    assert noteset.notes_octave_fit == tuple(notes_octave_fit)


@pytest.mark.parametrize(
    'noteset, note, steps, result', (
        (NoteSet.from_str('CDEFGAB'), Note('C'), 3, Note('F')),
        (NoteSet.from_str('CDEFGAB'), Note('C'), -2, Note('A')),
        (NoteSet.from_str('DEFGAbC'), Note('A'), 1, Note('b')),
        (NoteSet.from_str('DEFGAbC'), Note('A'), 0, Note('A')),
    ),
)
def test_add_note_abstract(noteset, note, steps, result):
    assert noteset.add_note(note, steps) == result


@pytest.mark.parametrize(
    'noteset, note, steps', (
        (NoteSet(frozenset()), 'A', 1),
        (NoteSet(frozenset()), 'A1', 1),
    ),
)
def test_add_note_empty_noteset(noteset, note, steps):
    with pytest.raises(NotImplementedError):
        noteset.add_note(note, steps)


def _make_keyboard(notes: Sequence[Note], octaves: Sequence[int]) -> tuple[SpecificNote, ...]:
    return tuple(sorted(SpecificNote(note, octave) for octave, note in itertools.product(octaves, notes)))


def _add_note_specific_generator():
    notesets = [NoteSet.from_str('CDEFGAB'), NoteSet.from_str('DEFGAbC')]
    notesets += list(all_scales['diatonic'].values())
    notesets += [
        Scale.from_name('C', 'h_minor'),
        Scale.from_name('E', 'h_minor'),
        Scale.from_name('d', 'm_minor'),
        Scale.from_name('f', 's_minor'),
        Scale.from_name('b', 'p_minor'),
    ]
    for noteset in notesets:
        noteset_str = f'NoteSet({noteset})'
        if isinstance(noteset, Scale):
            noteset_str = f'Scale({noteset.root.name}, {noteset.name})'
        yield pytest.param(noteset, id=noteset_str)


@pytest.mark.parametrize('noteset', _add_note_specific_generator())
def test_add_note_specific(noteset):
    keyboard = _make_keyboard(notes=noteset.notes_ascending, octaves=range(-10, 10))
    for note, octave, steps in itertools.product(
        [noteset.notes_ascending[0], noteset.notes_ascending[1], noteset.notes_ascending[2], noteset.notes_ascending[-1]],
        [-2, -1, 0, 1, 2],
        [-29, -13, -8, -7, -6, -2, -1, 0, 1, 2, 6, 7, 8, 13, 29],
    ):
        note = SpecificNote(note, octave)
        result = keyboard[keyboard.index(note) + steps]
        assert noteset.add_note(note, steps) == result


@pytest.mark.parametrize(
    'noteset, root, expected', (
        ('CD/C', 'C', 'CD/C'),
        ('CD/C', 'D', 'CD/D'),
        ('CD/D', 'C', 'CD/C'),
        ('CD', 'C', 'CD/C'),
    ),
)
def test_with_root(noteset, root, expected):
    assert NoteSet.from_str(noteset).with_root(Note(root)) == NoteSet.from_str(expected)


def test_with_root_validation():
    with pytest.raises(NotImplementedError):
        NoteSet(frozenset()).with_root(Note('C'))
    with pytest.raises(KeyError):
        NoteSet.from_str('CD/D').with_root(Note('E'))


@pytest.mark.parametrize(
    'noteset, note, expected', (
        ('CDEFGAB/C', 'A', 'ABdDEfa/A'),
        ('CdeFGab/e', 'D', 'DEfGABC/D'),
        ('Cd/C', 'd', 'dD/d'),
    ),
)
def test_transpose_to(noteset, note, expected):
    assert NoteSet.from_str(noteset).transpose_to(note) is NoteSet.from_str(expected)


@pytest.mark.parametrize(
    'notes, left, right, distance', (
        ('CDEFGAB', Note('E'), Note('C'), 2),
        ('CDEFGAB', Note('C'), Note('E'), 5),
        ('CDEFGAB', Note('B'), Note('C'), 6),
        ('CDEFGAB', Note('C'), Note('C'), 0),
        ('CDEFGAB', Note('E'), Note('A'), 4),
        ('CDE', Note('D'), Note('D'), 0),
        ('CDE', Note('E'), Note('D'), 1),
        ('CDE', Note('E'), Note('C'), 2),
        ('CDE', Note('C'), Note('D'), 2),
        ('CDE', Note('C'), Note('E'), 1),
        ('ab', Note('a'), Note('a'), 0),
        ('ab', Note('a'), Note('b'), 1),
        ('ab', Note('b'), Note('a'), 1),
        ('f', Note('f'), Note('f'), 0),
        ('CdDeEFfGaAbB', Note('b'), Note('b'), 0),
        ('CdDeEFfGaAbB', Note('G'), Note('C'), 7),
        ('CdDeEFfGaAbB', Note('C'), Note('d'), 11),
        ('CdDeEFfGaAbB', Note('C'), Note('G'), 5),

        ('CDEFGAB', SpecificNote('E', 1), SpecificNote('C', 1), 2),
        ('CDEFGAB', SpecificNote('E', 3), SpecificNote('C', 1), 16),
        ('CDEFGAB', SpecificNote('C', 1), SpecificNote('E', 3), -16),
        ('CDEFGAB', SpecificNote('C', 2), SpecificNote('E', -1), 19),
        ('CDEFGAB', SpecificNote('E', -1), SpecificNote('C', 2), -19),
        ('CDEFGAB', SpecificNote('C', 2), SpecificNote('E', -3), 33),
        ('CDEFGAB', SpecificNote('E', -3), SpecificNote('C', 2), -33),
        ('CDEFGAB', SpecificNote('C', -2), SpecificNote('E', -3), 5),
        ('CDEFGAB', SpecificNote('E', -3), SpecificNote('C', -2), -5),
        ('CDEFGAB', SpecificNote('B', 1), SpecificNote('C', 1), 6),
        ('CDEFGAB', SpecificNote('C', 1), SpecificNote('B', 1), -6),
        ('CDEFGAB', SpecificNote('B', 10), SpecificNote('C', 1), 69),
        ('CDEFGAB', SpecificNote('C', 1), SpecificNote('B', 10), -69),
        ('CDEFGAB', SpecificNote('C', 0), SpecificNote('C', 0), 0),
        ('CDEFGAB', SpecificNote('F', 34), SpecificNote('F', 34), 0),
        ('CDEFGAB', SpecificNote('E', 4), SpecificNote('A', 2), 11),
        ('CDEFGAB', SpecificNote('A', 2), SpecificNote('E', 4), -11),
        ('CDE', SpecificNote('D', 2), SpecificNote('D', 2), 0),
        ('CDE', SpecificNote('D', 2), SpecificNote('D', 3), -3),
        ('CDE', SpecificNote('E', 5), SpecificNote('D', 4), 4),
        ('CDE', SpecificNote('D', 4), SpecificNote('E', 5), -4),
        ('CDE', SpecificNote('E', 5), SpecificNote('C', 4), 5),
        ('CDE', SpecificNote('C', 4), SpecificNote('E', 5), -5),
        ('ab', SpecificNote('a', 3), SpecificNote('a', 3), 0),
        ('ab', SpecificNote('b', 3), SpecificNote('a', 3), 1),
        ('ab', SpecificNote('a', 3), SpecificNote('b', 3), -1),
        ('ab', SpecificNote('b', 4), SpecificNote('a', 3), 3),
        ('ab', SpecificNote('a', 3), SpecificNote('b', 4), -3),
        ('f', SpecificNote('f', 0), SpecificNote('f', 0), 0),
        ('f', SpecificNote('f', 1), SpecificNote('f', 0), 1),
        ('f', SpecificNote('f', 0), SpecificNote('f', 1), -1),
        ('f', SpecificNote('f', 2), SpecificNote('f', 0), 2),
        ('f', SpecificNote('f', 0), SpecificNote('f', 2), -2),
        ('f', SpecificNote('f', 40), SpecificNote('f', 1), 39),
        ('f', SpecificNote('f', 1), SpecificNote('f', 40), -39),
        ('f', SpecificNote('f', 1), SpecificNote('f', -2), 3),
        ('f', SpecificNote('f', -2), SpecificNote('f', 1), -3),
        ('f', SpecificNote('f', -4), SpecificNote('f', -7), 3),
        ('f', SpecificNote('f', -7), SpecificNote('f', -4), -3),
        ('CdDeEFfGaAbB', SpecificNote('b', 2), SpecificNote('b', 2), 0),
        ('CdDeEFfGaAbB', SpecificNote('G', 5), SpecificNote('C', 3), 31),
        ('CdDeEFfGaAbB', SpecificNote('C', 3), SpecificNote('G', 5), -31),
        ('CdDeEFfGaAbB', SpecificNote('C', 2), SpecificNote('d', -1), 35),
        ('CdDeEFfGaAbB', SpecificNote('d', -1), SpecificNote('C', 2), -35),
        ('CdDeEFfGaAbB', SpecificNote('C', -2), SpecificNote('C', -3), 12),
        ('CdDeEFfGaAbB', SpecificNote('C', -3), SpecificNote('C', -2), -12),
        ('CdDeEFfGaAbB', SpecificNote('C', -3), SpecificNote('C', -8), 60),
        ('CdDeEFfGaAbB', SpecificNote('C', -8), SpecificNote('C', -3), -60),
        ('CdDeEFfGaAbB', SpecificNote('d', -3), SpecificNote('G', -8), 54),
        ('CdDeEFfGaAbB', SpecificNote('G', -8), SpecificNote('d', -3), -54),
    ),
)
def test_subtract(notes, left, right, distance):
    assert NoteSet.from_str(notes).subtract(left, right) == distance


def test_subtract_types():
    noteset = NoteSet.from_str('CDEFGAB')
    with pytest.raises(TypeError):
        noteset.subtract(Note('C'), SpecificNote('D', 1))
    with pytest.raises(TypeError):
        noteset.subtract(SpecificNote('D', 1), Note('C'))
    with pytest.raises(TypeError):
        noteset.subtract('C', SpecificNote('D', 1))  # type: ignore
    with pytest.raises(TypeError):
        noteset.subtract('D1', Note('C'))  # type: ignore



import operator

import pytest

from musiclib.chord import Chord
from musiclib.chord import SpecificChord
from musiclib.note import Note
from musiclib.note import SpecificNote
from musiclib.noteset import NoteSet
from musiclib.progression import Progression
from musiclib.scale import Scale
from musiclib.util.cache import Cached


@pytest.mark.parametrize(
    'op, a, b', (
        (operator.is_, Note('A'), Note('A')),
        (operator.is_not, Note('A'), Note('B')),
    ),
)
def test_note(op, a, b):
    assert op(a, b)


@pytest.mark.parametrize(
    'op, a, b', (
        (operator.is_, SpecificNote('C', 1), SpecificNote('C', 1)),
        (operator.is_not, SpecificNote('C', 1), SpecificNote('C', 2)),
        (operator.is_, SpecificNote('C', 1), SpecificNote(Note('C'), 1)),
    ),
)
def test_specific_note(op, a, b):
    assert op(a, b)


@pytest.mark.parametrize(
    'op, a, b', (
        (operator.is_, NoteSet(frozenset(map(Note, 'CDe'))), NoteSet(frozenset(map(Note, 'CDe')))),
        (operator.is_, NoteSet(frozenset(map(Note, 'CDe'))), NoteSet.from_str('CDe')),
        (operator.is_, NoteSet(frozenset(map(Note, 'CDe')), root=Note('C')), NoteSet(frozenset(map(Note, 'CDe')), root=Note('C'))),
        (operator.is_, NoteSet(frozenset(map(Note, 'CDe')), root=Note('C')), NoteSet.from_str('CDe/C')),
        (operator.is_, NoteSet.from_str('CDe/C'), NoteSet.from_str('CDe/C')),
        (operator.is_not, NoteSet.from_str('CDe/C'), NoteSet.from_str('CDe/D')),
    ),
)
def test_noteset(op, a, b):
    assert op(a, b)


@pytest.mark.parametrize(
    'op, a, b', (
        (operator.is_, Scale.from_str('CdeFGab/C'), Scale.from_str('CdeFGab/C')),
        (operator.is_, Scale(frozenset(map(Note, 'CdeFGab')), root=Note('C')), Scale.from_str('CdeFGab/C')),
        (operator.is_not, Scale.from_str('CdeFGab/C'), Scale.from_str('CdeFGab/d')),
        (operator.is_, Scale.from_name('C', 'major'), Scale.from_name('C', 'major')),
        (operator.is_not, Scale.from_name('C', 'major'), Scale.from_name('D', 'major')),
        (operator.is_, Scale.from_name('C', 'major'), Scale.from_name(Note('C'), 'major')),
        (operator.is_, Scale.from_name('C', 'major'), Scale.from_str('CDEFGAB/C')),
    ),
)
def test_scale(op, a, b):
    assert op(a, b)


@pytest.mark.parametrize(
    'op, a, b', (
        (operator.is_, Chord(frozenset(map(Note, 'CEG')), root=Note('C')), Chord(frozenset(map(Note, 'CEG')), root=Note('C'))),
        (operator.is_, Chord(frozenset(map(Note, 'CEG')), root=Note('C')), Chord.from_str('CEG/C')),
        (operator.is_, Chord.from_str('CEG/C'), Chord.from_str('CEG/C')),
        (operator.is_not, Chord.from_str('CEG/C'), Chord.from_str('CeG/C')),
        (operator.is_not, Chord.from_str('CEG/C'), Chord.from_str('CEG/E')),
        (operator.is_, Chord.from_name('C', 'major'), Chord.from_name('C', 'major')),
        (operator.is_not, Chord.from_name('C', 'major'), Chord.from_name('D', 'major')),
        (operator.is_, Chord.from_name('C', 'major'), Chord.from_name(Note('C'), 'major')),
    ),
)
def test_chord(op, a, b):
    assert op(a, b)


def test_specific_chord():
    a = SpecificNote('C', 5)
    b = SpecificNote('E', 5)
    c = SpecificNote('G', 5)
    d = SpecificNote('C', 6)
    assert SpecificChord(frozenset({a, b, c})) is SpecificChord(frozenset({a, b, c}))
    assert SpecificChord(frozenset({a, b, c}), root='C') is SpecificChord(frozenset({a, b, c}), root=Note('C'))
    assert SpecificChord(frozenset({a, b, c}), root='C') is not SpecificChord(frozenset({a, b, d}), root=Note('C'))
    assert SpecificChord(frozenset({a, b, c}), root='C') is not SpecificChord(frozenset({a, b, c}), root=Note('E'))
    assert SpecificChord.from_str('C1_E1_G1/C') is SpecificChord.from_str('C1_E1_G1/C')
    assert SpecificChord.from_str('C1_E1_G1/C') is not SpecificChord.from_str('C1_E1_G1/E')
    assert SpecificChord.from_str('C1_E1_G1/C') is not SpecificChord.from_str('C1_e1_G1/C')


def test_progression():
    p0 = Progression((
        SpecificChord.from_str('G2_B2_e3'),
        SpecificChord.from_str('A2_C3_E3'),
        SpecificChord.from_str('B2_D3_f3'),
        SpecificChord.from_str('C3_E3_G3'),
    ))
    p1 = Progression((
        SpecificChord.from_str('G2_B2_e3'),
        SpecificChord.from_str('A2_C3_E3'),
        SpecificChord.from_str('B2_D3_f3'),
        SpecificChord.from_str('C3_E3_G3'),
    ))
    p2 = Progression((
        SpecificChord.from_str('C0_E0_a0'),
        SpecificChord.from_str('D0_F0_A0'),
        SpecificChord.from_str('E0_G0_B0'),
        SpecificChord.from_str('F0_A0_C1'),
    ))
    assert p0 is p1
    assert p0 is not p2


def test_cached_class():
    class K(Cached):
        def __init__(self, x):
            self.x = x

        def __bool__(self):
            """
            test walrus operator is properly used
            explicit `is not None` check should be used instead of `if cached := cache.get(key):`
            """
            return False

    assert K(1) is K(1)
    assert K(1) is not K(2)



from collections.abc import Sequence

import pytest

from musiclib import config
from musiclib.chord import Chord
from musiclib.note import SpecificNote
from musiclib.noterange import NoteRange
from musiclib.noteset import NoteSet
from musiclib.scale import Scale


@pytest.mark.parametrize(
    'start, stop, noteset, expected', (
        ('C0', 'C1', NoteSet.from_str(config.chromatic_notes), 'C0 d0 D0 e0 E0 F0 f0 G0 a0 A0 b0 B0 C1'),
        ('b3', 'E4', NoteSet.from_str(config.chromatic_notes), 'b3 B3 C4 d4 D4 e4 E4'),
        ('C0', 'C0', NoteSet.from_str(config.chromatic_notes), 'C0'),
        ('C0', 'C1', NoteSet.from_str('CDEFGAB'), 'C0 D0 E0 F0 G0 A0 B0 C1'),
        ('C0', 'C1', Scale.from_str('CDEFGAB/C'), 'C0 D0 E0 F0 G0 A0 B0 C1'),
        ('C0', 'C1', Chord.from_str('CDEFGAB/C'), 'C0 D0 E0 F0 G0 A0 B0 C1'),
        ('a3', 'f4', NoteSet.from_str('dEfaB'), 'a3 B3 d4 E4 f4'),
        ('A0', 'D2', NoteSet.from_str('CDEFGAB'), 'A0 B0 C1 D1 E1 F1 G1 A1 B1 C2 D2'),
    ),
)
def test_note_range(start, stop, noteset, expected):
    assert list(NoteRange(SpecificNote.from_str(start), SpecificNote.from_str(stop), noteset)) == [SpecificNote.from_str(s) for s in expected.split()]


@pytest.mark.parametrize(
    'start, stop, noterange', (
        ('C0', 'C1', NoteRange(SpecificNote('C', 0), SpecificNote('C', 1))),
        ('E1', 'f3', NoteRange(SpecificNote('E', 1), SpecificNote('f', 3))),
        (SpecificNote('E', 1), 'f3', NoteRange(SpecificNote('E', 1), SpecificNote('f', 3))),
        ('E1', SpecificNote('f', 3), NoteRange(SpecificNote('E', 1), SpecificNote('f', 3))),
        (SpecificNote('E', 1), SpecificNote('f', 3), NoteRange(SpecificNote('E', 1), SpecificNote('f', 3))),
    ),
)
def test_note_range_from_str(start, stop, noterange):
    assert NoteRange(start, stop) == noterange


def test_noterange_bounds():
    with pytest.raises(ValueError):
        NoteRange(SpecificNote('D', 2), SpecificNote('C', 1))
    with pytest.raises(KeyError):
        NoteRange(SpecificNote('C', 1), SpecificNote('C', 2), noteset=NoteSet(frozenset()))
    with pytest.raises(KeyError):
        NoteRange(SpecificNote('C', 1), SpecificNote('D', 2), noteset=NoteSet.from_str('Cd'))
    with pytest.raises(KeyError):
        NoteRange(SpecificNote('C', 1), SpecificNote('D', 2), noteset=NoteSet.from_str('dD'))
    with pytest.raises(KeyError):
        NoteRange(SpecificNote('C', 1), SpecificNote('D', 2), noteset=NoteSet.from_str('dDeE'))


def test_noterange_contains():
    assert SpecificNote('D', 1) in NoteRange(SpecificNote('C', 1), SpecificNote('C', 2))
    assert SpecificNote('C', 1) in NoteRange(SpecificNote('C', 1), SpecificNote('C', 2))
    assert SpecificNote('C', 2) in NoteRange(SpecificNote('C', 1), SpecificNote('C', 2))
    assert SpecificNote('C', 3) not in NoteRange(SpecificNote('C', 1), SpecificNote('C', 2))
    assert SpecificNote('D', 1) not in NoteRange(SpecificNote('C', 1), SpecificNote('F', 1), noteset=NoteSet.from_str('CEF'))


@pytest.mark.parametrize(
    'start, stop, notes, length', (
        (SpecificNote('C', 1), SpecificNote('G', 1), 'CdDeEFfGaAbB', 8),
        (SpecificNote('D', 1), SpecificNote('G', 3), 'CdDeEFfGaAbB', 30),
        (SpecificNote('D', 1), SpecificNote('D', 1), 'CdDeEFfGaAbB', 1),
        (SpecificNote('E', 1), SpecificNote('b', 1), 'CDEFGAb', 5),
        (SpecificNote('b', 1), SpecificNote('G', 3), 'CDEFGAb', 13),
        (SpecificNote('f', 1), SpecificNote('a', 1), 'fa', 2),
        (SpecificNote('f', 1), SpecificNote('f', 2), 'fa', 3),
        (SpecificNote('f', 1), SpecificNote('a', 3), 'fa', 6),
        (SpecificNote('f', 1), SpecificNote('f', 3), 'f', 3),
    ),
)
def test_noterange_len(start, stop, notes, length):
    assert len(NoteRange(start, stop, NoteSet.from_str(notes))) == length


def test_noterange_getitem():
    nr = NoteRange(SpecificNote('C', 1), SpecificNote('C', 2))
    assert nr[0] == nr[-13] == SpecificNote('C', 1)
    assert nr[1] == nr[-12] == SpecificNote('d', 1)
    assert nr[2] == nr[-11] == SpecificNote('D', 1)
    assert nr[12] == nr[-1] == SpecificNote('C', 2)
    assert nr[11] == nr[-2] == SpecificNote('B', 1)
    assert nr[0:0] == NoteRange(SpecificNote('C', 1), SpecificNote('C', 1))
    assert nr[0:1] == NoteRange(SpecificNote('C', 1), SpecificNote('d', 1))
    assert nr[0:2] == NoteRange(SpecificNote('C', 1), SpecificNote('D', 1))
    assert nr[0:12] == NoteRange(SpecificNote('C', 1), SpecificNote('C', 2))
    assert nr[1:] == NoteRange('d1', 'C2')
    assert nr[:4] == NoteRange('C1', 'E1')
    assert nr[:] == NoteRange('C1', 'C2')

    with pytest.raises(IndexError):
        nr[13]
    with pytest.raises(IndexError):
        nr[-14]
    with pytest.raises(IndexError):
        nr[-3: 1]
    with pytest.raises(IndexError):
        nr[5: 13]

    ns = NoteSet.from_str('fa')
    nr = NoteRange(SpecificNote('f', -1), SpecificNote('a', 3), ns)
    assert nr[0] == nr[-10] == SpecificNote('f', -1)
    assert nr[1] == nr[-9] == SpecificNote('a', -1)
    assert nr[2] == nr[-8] == SpecificNote('f', 0)
    assert nr[9] == nr[-1] == SpecificNote('a', 3)
    assert nr[8] == nr[-2] == SpecificNote('f', 3)
    assert nr[0:0] == NoteRange(SpecificNote('f', -1), SpecificNote('f', -1), ns)
    assert nr[0:1] == NoteRange(SpecificNote('f', -1), SpecificNote('a', -1), ns)
    assert nr[0:2] == NoteRange(SpecificNote('f', -1), SpecificNote('f', 0), ns)
    assert nr[0:9] == NoteRange(SpecificNote('f', -1), SpecificNote('a', 3), ns)
    with pytest.raises(IndexError):
        nr[10]
    with pytest.raises(IndexError):
        nr[-11]


@pytest.mark.parametrize(
    'noterange, expected', [
        (NoteRange(SpecificNote('C', 1), SpecificNote('C', 2)), 'C1 d1 D1 e1 E1 F1 f1 G1 a1 A1 b1 B1 C2'),
        (NoteRange(SpecificNote('b', 1), SpecificNote('D', 2), noteset=NoteSet.from_str('AbBCdDe')), 'b1 B1 C2 d2 D2'),
    ],
)
def test_noterange_list(noterange, expected):
    assert list(noterange) == [SpecificNote.from_str(s) for s in expected.split()]


def test_sequence():
    nr = NoteRange(SpecificNote('C', 1), SpecificNote('C', 2))
    assert isinstance(nr, Sequence)



from collections import deque

import pytest

from musiclib.rhythm import Rhythm


@pytest.fixture
def example_notes():
    return 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0


@pytest.mark.parametrize('n_notes', range(1, 16 + 1))
def test_n_notes(n_notes):
    assert sum(Rhythm.random_rhythm(n_notes).notes) == n_notes


@pytest.mark.parametrize('n_notes', (-1, 16 + 1))
def test_n_notes_validation(n_notes):
    with pytest.raises(ValueError):
        Rhythm.random_rhythm(n_notes)


@pytest.mark.parametrize(
    'notes, score', [
        ((0, 0, 1, 1, 0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1), 0.3),
        ((1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1), 0.5),
        ((0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0), 12.5),
        ((1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0), 1.5833333333333333),
        ((1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 0), 3),
        ((1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0), 18),
        ((0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0), 2.3333333333333335),
        ((0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 1), 6.333333333333333),
        ((1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1), float('inf')),
    ],
)
def test_score(notes, score):
    assert Rhythm(notes, bar_notes=16).score == score


def test_rhythm_score(example_notes):
    more_notes = 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    assert Rhythm(example_notes).score < Rhythm(more_notes).score


def test_score_rotation(example_notes):
    original_score = Rhythm(example_notes).score
    rotated_notes = deque(example_notes)
    for _ in range(len(rotated_notes)):
        rotated_notes.rotate(1)
        assert Rhythm(tuple(rotated_notes)).score == original_score


def test_has_contiguous_ones(example_notes):
    assert Rhythm(example_notes).has_contiguous_ones
    assert Rhythm((1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1)).has_contiguous_ones
    assert not Rhythm((1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0)).has_contiguous_ones


def test_all_rhythms():
    assert Rhythm.all_rhythms(
        n_notes=3,
        bar_notes=6,
    ) == (
        Rhythm((0, 1, 0, 1, 0, 1), bar_notes=6),
        Rhythm((1, 0, 0, 1, 0, 1), bar_notes=6),
        Rhythm((1, 0, 1, 0, 0, 1), bar_notes=6),
        Rhythm((1, 0, 1, 0, 1, 0), bar_notes=6),
    )



import random

import pytest

from musiclib.chord import Chord
from musiclib.chord import SpecificChord
from musiclib.note import Note
from musiclib.note import SpecificNote
from musiclib.noteset import NoteSet


def test_creation_from_notes():
    assert str(Chord(frozenset({Note('C'), Note('E'), Note('G')}), root=Note('C'))) == 'CEG/C'


def test_chord_root_validation():
    with pytest.raises(TypeError):
        Chord.from_str('CEG')


@pytest.mark.parametrize('arg', ('C1_D1_E1', set('C1_D1_E1'), tuple('C1_D1_E1'), list('C1_D1_E1')))
def test_specificchord_notes_type_is_frozenset(arg):
    with pytest.raises(TypeError):
        SpecificChord(arg)


def test_specificchord_root_validation():
    assert isinstance(SpecificChord(frozenset({SpecificNote('A', 1)}), root='A').root, Note)
    with pytest.raises(KeyError):
        SpecificChord(frozenset({SpecificNote('A', 1)}), root='E')


def test_notes():
    assert Chord(frozenset(map(Note, 'CEG')), root=Note('C')).notes == frozenset({Note('C'), Note('E'), Note('G')})
    assert Chord.from_name('C', 'major').notes == frozenset({Note('C'), Note('E'), Note('G')})


@pytest.mark.parametrize(
    'chord, expected', (
        (Chord(frozenset(map(Note, 'BDF')), root=Note('B')), 'BDF/B'),
        (Chord(frozenset(map(Note, 'DGBFCEA')), root=Note('B')), 'BCDEFGA/B'),
    ),
)
def test_str_sort_2_octaves(chord, expected):
    assert str(chord) == expected


@pytest.mark.parametrize(
    'notes, root, expected', (
        ('CEG', 'C', 'major'),
        ('BDF', 'B', 'diminished'),
        ('CefA', 'C', 'dim7'),
        ('DFaC', 'D', 'half-dim7'),
    ),
)
def test_name(notes, root, expected):
    assert Chord(frozenset(map(Note, notes)), root=Note(root)).name == expected


def test_intervals():
    assert Chord.from_str('CEG/C').intervals == frozenset({0, 4, 7})


def test_from_name():
    assert str(Chord.from_name('C', 'major')) == 'CEG/C'
    assert str(Chord.from_name('d', '7')) == 'dFaB/d'


def test_from_str():
    assert SpecificChord.from_str('C1_E1_G1/C') == SpecificChord(frozenset({SpecificNote('C', 1), SpecificNote('E', 1), SpecificNote('G', 1)}), root=Note('C'))
    assert Chord.from_str('CEG/C') == Chord(frozenset(map(Note, 'CEG')), root=Note('C'))
    for _ in range(10):
        s_chord = SpecificChord.random()
        assert SpecificChord.from_str(str(s_chord)) == s_chord
        abstract_notes = SpecificNote.to_abstract(s_chord.notes)
        chord = Chord(notes=abstract_notes, root=random.choice(tuple(abstract_notes)))
        assert Chord.from_str(str(chord)) == chord
    with pytest.raises(NotImplementedError):
        SpecificChord.from_str('C1_C1')
    with pytest.raises(NotImplementedError):
        SpecificChord.from_str('C1_E1_E1')


def test_magic_methods():
    chord = SpecificChord.from_str('C1_E1_G1')
    assert len(chord) == len(chord.notes)
    assert all(chord[i] == chord.notes_ascending[i] for i in range(len(chord)))
    assert tuple(iter(chord)) == chord.notes_ascending
    chord2 = SpecificChord.from_str('d3_f8')
    assert tuple(zip(chord, chord2)) == (
        (SpecificNote('C', 1), SpecificNote('d', 3)),
        (SpecificNote('E', 1), SpecificNote('f', 8)),
    )


def test_sub():
    a = SpecificChord.from_str('C1_E1_G1')
    b = SpecificChord.from_str('B0_D1_G1')
    assert a - b == b - a == 3


def test_combinations_order():
    for _ in range(10):
        for n, m in SpecificChord.random().notes_combinations():
            assert n < m


@pytest.mark.parametrize(
    'specific, abstract', (
        ('C1_E1_G2/C', Chord.from_str('CEG/C')),
        ('C1_E1_G2', NoteSet.from_str('CEG')),
    ),
)
def test_abstract(specific, abstract):
    assert SpecificChord.from_str(specific).abstract == abstract


def test_find_intervals():
    a = SpecificNote('C', 5)
    b = SpecificNote('E', 5)
    c = SpecificNote('G', 5)
    d = SpecificNote('C', 6)
    e = SpecificNote('D', 6)

    assert SpecificChord(frozenset({a, b, c})).find_intervals(7) == ((a, c),)
    assert SpecificChord(frozenset({a, b, c, e})).find_intervals(7) == ((a, c), (c, e))
    assert SpecificChord(frozenset({a, d})).find_intervals(12) == ((a, d),)


@pytest.mark.parametrize(
    'chord, note, steps, result', (
        (Chord.from_str('CEG/C'), Note('C'), 1, Note('E')),
        (Chord.from_str('CEG/C'), Note('C'), 2, Note('G')),
        (Chord.from_str('CEG/C'), Note('C'), 3, Note('C')),
        (Chord.from_str('CeGb/C'), Note('e'), 2, Note('b')),
        (Chord.from_str('CeGb/C'), Note('e'), 25, Note('G')),
        (Chord.from_str('CEG/C'), Note('C'), -1, Note('G')),
        (Chord.from_str('CEG/C'), Note('C'), -2, Note('E')),
        (Chord.from_str('CEG/C'), Note('C'), -3, Note('C')),
        (Chord.from_str('CEG/C'), Note('C'), -3, Note('C')),
        (Chord.from_str('CeGb/C'), Note('e'), -15, Note('G')),
    ),
)
def test_add_note(chord, note, steps, result):
    assert chord.add_note(note, steps) == result


@pytest.mark.parametrize(
    'chord, expected', (
        ('C3_E3_G3', 'C0_E0_G0'),
        ('F3_A3_C4', 'C0_E0_G0'),
        ('C3_E3_G3/C', 'C0_E0_G0/C'),
        ('F3_A3_C4/F', 'C0_E0_G0/C'),
    ),
)
def test_transposed_to_C0(chord, expected):
    assert SpecificChord.from_str(chord).transposed_to_C0 == SpecificChord.from_str(expected)


@pytest.mark.parametrize(
    'chord, add, expected', (
        ('C3_E3_G3', -36, 'C0_E0_G0'),
        ('C3_E3_G3', -24, 'C1_E1_G1'),
        ('C3_E3_G3', 0, 'C3_E3_G3'),
        ('C3_E3_G3', 12, 'C4_E4_G4'),
        ('C3_E3_G3', -34, 'D0_f0_A0'),
        ('C3_E3_G3', 13, 'd4_F4_a4'),
        ('C3_E3_G3/C', -36, 'C0_E0_G0/C'),
        ('C3_E3_G3/C', -24, 'C1_E1_G1/C'),
        ('C3_E3_G3/C', 13, 'd4_F4_a4/d'),
        ('C3_E3_G3/E', 13, 'd4_F4_a4/F'),
    ),
)
def test_add(chord, add, expected):
    assert SpecificChord.from_str(chord) + add == SpecificChord.from_str(expected)
    with pytest.raises(TypeError):
        chord + [1]



import pickle

import pytest

from musiclib.chord import Chord
from musiclib.chord import SpecificChord
from musiclib.note import Note
from musiclib.note import SpecificNote
from musiclib.noteset import NoteSet
from musiclib.progression import Progression
from musiclib.scale import Scale


@pytest.mark.parametrize('note', (Note('C'), SpecificNote('C', 3)))
def test_note(note):
    assert note == pickle.loads(pickle.dumps(note))


def iter_containers(a, b):
    yield from (
        (a, b),
        [a, b],
        {a, b},
        frozenset({a, b}),
        {a: 1, b: 1},
        {1: a, 2: b},
    )


@pytest.mark.parametrize(
    'container', (
        {1: Note('C')},
        *iter_containers(Note('C'), Note('D')),
        *iter_containers(SpecificNote('C', 3), SpecificNote('D', 3)),
        *iter_containers(SpecificNote('C', 3), SpecificNote('C', 4)),
        *iter_containers(SpecificNote('C', 3), SpecificNote('D', 4)),
    ),
)
def test_note_container(container):
    assert container == pickle.loads(pickle.dumps(container))


a = NoteSet.from_str('fa/a')
b = NoteSet.from_str('fa/f')


@pytest.mark.parametrize(
    'noteset', (
        NoteSet.from_str('CDEFGAB/C'),
        NoteSet.from_str('CDEFGAB'),
        NoteSet.from_str('CdeFGab/e'),
        NoteSet.from_str('CEG/C'),
        a,
        *iter_containers(a, b),
    ),
)
def test_noteset(noteset):
    assert noteset == pickle.loads(pickle.dumps(noteset))


c = Chord(frozenset({Note('C'), Note('E'), Note('G')}), root=Note('C'))
d = Chord(frozenset({Note('C'), Note('E'), Note('G')}), root=Note('E'))
e = SpecificChord(frozenset({SpecificNote('C', 5), SpecificNote('E', 5), SpecificNote('G', 5)}), root=Note('C'))
f = SpecificChord(frozenset({SpecificNote('C', 5), SpecificNote('E', 5), SpecificNote('G', 5)}), root=Note('E'))


@pytest.mark.parametrize(
    'chord', (
        Chord(frozenset({Note('C'), Note('E'), Note('G')}), root=Note('C')),
        c,
        SpecificChord(frozenset({SpecificNote('C', 5), SpecificNote('E', 5), SpecificNote('G', 5)})),
        e,
        f,
        *iter_containers(c, d),
        *iter_containers(e, f),
    ),
)
def test_chord(chord):
    assert chord == pickle.loads(pickle.dumps(chord))


@pytest.mark.parametrize(
    'scale', (
        Scale.from_name('C', 'major'),
        Scale.from_str('DEFGAbC/D'),
        *iter_containers(Scale.from_name('C', 'major'), Scale.from_name('D', 'major')),
    ),
)
def test_scale(scale):
    assert scale == pickle.loads(pickle.dumps(scale))


g = SpecificChord.random()
h = SpecificChord.random()
i = SpecificChord.random()
j = SpecificChord.random()
k = SpecificChord.random()


@pytest.mark.parametrize(
    'progression', (
        Progression((g, h, i, j)),
        *iter_containers(Progression((g, h, i, j)), Progression((g, h, i, k))),
    ),
)
def test_progression(progression):
    assert progression == pickle.loads(pickle.dumps(progression))



import itertools

import pytest
from colortool import Color

from musiclib.chord import Chord
from musiclib.chord import SpecificChord
from musiclib.noterange import NoteRange
from musiclib.noteset import NoteSet
from musiclib.scale import ComparedScales
from musiclib.scale import Scale
from musiclib.scale import all_scales

TITLE = 'title_fUYsZHfC'
SUBTITLE = 'subtitle_EfrKTj'
TITLE_HREF = 'title_href_TUMhv'
BACKGROUND_COLOR = Color.from_hex(0xFACE45)


def svg_helper(html, classes, title, subtitle, title_href, background_color):
    classes = ' '.join(classes)
    assert f'class="{classes}"' in html
    for item, constant in zip(
        (title, subtitle, title_href, background_color),
        (TITLE, SUBTITLE, TITLE_HREF, BACKGROUND_COLOR),
    ):
        if item is None:
            assert constant not in html
        elif isinstance(item, Color):
            assert item.css_hex in html
        else:
            if constant not in html:
                print(html)
            assert constant in html


@pytest.mark.parametrize(
    'noteset', [
        NoteSet.from_str('CdeFGa'),
        NoteSet.from_str('CdeFGa/e'),
        NoteSet.from_str('fa'),
        NoteSet.from_str('fa/f'),
        NoteSet.from_str(''),
    ],
)
@pytest.mark.parametrize('title', (None, TITLE))
@pytest.mark.parametrize('subtitle', (None, SUBTITLE))
@pytest.mark.parametrize('title_href', (None, TITLE_HREF))
@pytest.mark.parametrize('background_color', (BACKGROUND_COLOR,))
def test_html_noteset(noteset, title, subtitle, title_href, background_color):
    if title is None and title_href is not None:
        pytest.skip('title_href requires title')
    classes = ('cls1', 'cls2')
    svg = noteset._repr_svg_(
        classes=classes,
        title=title,
        subtitle=subtitle,
        title_href=title_href,
        background_color=background_color,
    )
    svg_helper(svg, classes, title, subtitle, title_href, background_color)


@pytest.mark.parametrize('kind', ('diatonic', 'harmonic', 'melodic', 'pentatonic', 'sudu'))
@pytest.mark.parametrize('title', (None, TITLE))
@pytest.mark.parametrize('subtitle', (None, SUBTITLE))
@pytest.mark.parametrize('title_href', (None, TITLE_HREF))
@pytest.mark.parametrize('background_color', (BACKGROUND_COLOR,))
def test_html_scale(kind, title, subtitle, title_href, background_color):
    if title is None and title_href is not None:
        pytest.skip('title_href requires title')
    for scale in all_scales[kind].values():
        classes = (scale.name,)
        svg = scale._repr_svg_(
            classes=classes,
            title=title,
            subtitle=subtitle,
            title_href=title_href,
            background_color=background_color,
        )
        svg_helper(svg, classes, title, subtitle, title_href, background_color)


@pytest.mark.parametrize(
    'scale0, scale1', (
        (Scale.from_name('C', 'major'), Scale.from_name('f', 'phrygian')),
        (Scale.from_name('A', 'major'), Scale.from_name('f', 'phrygian')),
    ),
)
@pytest.mark.parametrize('title', (None, TITLE))
@pytest.mark.parametrize('subtitle', (None, SUBTITLE))
@pytest.mark.parametrize('title_href', (None, TITLE_HREF))
@pytest.mark.parametrize('background_color', (BACKGROUND_COLOR,))
def test_html_compared_scale(scale0, scale1, title, subtitle, title_href, background_color):
    if title is None and title_href is not None:
        pytest.skip('title_href requires title')
    classes = ('cls1', 'cls2')
    svg = ComparedScales(scale0, scale1)._repr_svg_(
        classes=classes,
        title=title,
        subtitle=subtitle,
        title_href=title_href,
        background_color=background_color,
    )
    svg_helper(svg, classes, title, subtitle, title_href, background_color)


@pytest.mark.parametrize('root, name', itertools.product('Cf', Chord.name_to_intervals))
@pytest.mark.parametrize('title', (None, TITLE))
@pytest.mark.parametrize('subtitle', (None, SUBTITLE))
@pytest.mark.parametrize('title_href', (None, TITLE_HREF))
@pytest.mark.parametrize('background_color', (BACKGROUND_COLOR,))
def test_html_chord(root, name, title, subtitle, title_href, background_color):
    if title is None and title_href is not None:
        pytest.skip('title_href requires title')
    classes = ('cls1', 'cls2')
    svg = Chord.from_name(root, name)._repr_svg_(
        classes=classes,
        title=title,
        subtitle=subtitle,
        title_href=title_href,
        background_color=background_color,
    )
    svg_helper(svg, classes, title, subtitle, title_href, background_color)


@pytest.mark.parametrize(
    'chord', [
        SpecificChord.from_str('C1_E1_f1'),
        SpecificChord.from_str('C1_d3_A5'),
        SpecificChord(frozenset()),
    ],
)
@pytest.mark.parametrize('title', (None, TITLE))
@pytest.mark.parametrize('subtitle', (None, SUBTITLE))
@pytest.mark.parametrize('title_href', (None, TITLE_HREF))
@pytest.mark.parametrize('background_color', (BACKGROUND_COLOR,))
def test_html_specific_chord(chord, title, subtitle, title_href, background_color):
    if title is None and title_href is not None:
        pytest.skip('title_href requires title')
    classes = ('cls1', 'cls2')
    svg = chord._repr_svg_(
        classes=classes,
        title=title,
        subtitle=subtitle,
        title_href=title_href,
        background_color=background_color,
    )
    svg_helper(svg, classes, title, subtitle, title_href, background_color)


@pytest.mark.parametrize(
    'noterange', [
        NoteRange('C2', 'C5'),
        NoteRange('D2', 'G2', noteset=NoteSet.from_str('CDG')),
    ],
)
@pytest.mark.parametrize('title', (None, TITLE))
@pytest.mark.parametrize('subtitle', (None, SUBTITLE))
@pytest.mark.parametrize('title_href', (None, TITLE_HREF))
@pytest.mark.parametrize('background_color', (BACKGROUND_COLOR,))
def test_html_noterange(noterange, title, subtitle, title_href, background_color):
    if title is None and title_href is not None:
        pytest.skip('title_href requires title')
    classes = ('cls1', 'cls2')
    svg = noterange._repr_svg_(
        classes=classes,
        title=title,
        subtitle=subtitle,
        title_href=title_href,
        background_color=background_color,
    )
    svg_helper(svg, classes, title, subtitle, title_href, background_color)



import operator

import hypothesis.strategies as st
import pytest
from hypothesis import given

from musiclib.note import Note
from musiclib.note import SpecificNote


@pytest.mark.parametrize(
    'i, expected', (
        (0, 'C'),
        (12, 'C'),
        (1, 'd'),
        (13, 'd'),
        (-1, 'B'),
        (-28, 'a'),
    ),
)
def test_from_i(i, expected):
    assert Note.from_i(i) == expected


def test_note_exists():
    with pytest.raises(KeyError):
        Note('Q')
    with pytest.raises(KeyError):
        Note('1')


@pytest.mark.parametrize(
    'op, a, b', (
        (operator.eq, Note('C'), 'C'),
        (operator.eq, Note('C'), Note('C')),
        (operator.ne, Note('C'), 'D'),
        (operator.ne, Note('C'), Note('D')),
        (operator.lt, Note('C'), 'D'),
        (operator.lt, Note('C'), Note('D')),
        (operator.gt, Note('D'), Note('d')),
        (operator.gt, Note('B'), Note('f')),
        (operator.ne, SpecificNote.from_str('C1'), 'C1'),  # avoid implicit string comparison for SpecificNote
        (operator.ne, SpecificNote.from_str('C1'), 'C2'),
    ),
)
def test_ordering(op, a, b):
    assert op(a, b)


@given(st.integers())
def test_specific_note_from_i(i):
    assert SpecificNote.from_i(i).i == i


@pytest.mark.parametrize(
    'string, expected', (
        ('F1', SpecificNote('F', 1)),
        ('C4', SpecificNote('C', 4)),
        ('C10', SpecificNote('C', 10)),
        ('d456', SpecificNote('d', 456)),
        ('d-456', SpecificNote('d', -456)),
    ),
)
def test_from_str(string, expected):
    assert SpecificNote.from_str(string) == expected


@pytest.mark.parametrize(
    'note, steps, expected', (
        (Note('C'), 4, Note('E')),
        (Note('C'), 15, Note('e')),
        (Note('B'), 2, Note('d')),
        (Note('D'), -2, Note('C')),
        (Note('d'), -2, Note('B')),
    ),
)
def test_note_add(note, steps, expected):
    assert note + steps == expected


@given(st.integers(), st.integers())
def test_specific_note_add(i, to_add):
    note = SpecificNote.from_i(i)
    assert (note + to_add).i == note.i + to_add


@given(st.integers(), st.integers())
def test_specific_note_sub(i, to_sub):
    note = SpecificNote.from_i(i)
    assert note - to_sub == SpecificNote.from_i(note.i - to_sub)


def test_color():
    assert not Note('C').is_black
    assert Note('d').is_black
    assert not SpecificNote('D', 1).is_black
    assert SpecificNote('f', -35).is_black


def test_to_abstract():
    assert SpecificNote.to_abstract({SpecificNote('C', 2), SpecificNote('G', 3)}) == frozenset({Note('C'), Note('G')})


def test_midi_code():
    A4 = SpecificNote(Note('A'), 4)
    assert A4.i == 69
    assert SpecificNote.from_i(69) == A4



from collections.abc import Sequence

import pytest

from musiclib.chord import SpecificChord
from musiclib.progression import Progression


@pytest.fixture
def four_chords():
    a = SpecificChord.random()
    b = SpecificChord.random()
    c = SpecificChord.random()
    d = SpecificChord.random()
    return a, b, c, d


@pytest.fixture
def progression4():
    a = SpecificChord.from_str('C1_E1_G1')
    b = SpecificChord.from_str('D1_F1_A1')
    c = SpecificChord.from_str('E1_G1_B1')
    d = SpecificChord.from_str('F1_A1_C2')
    return Progression((a, b, c, d))


def test_validation():
    with pytest.raises(TypeError):
        Progression((0, 1, 2))  # type: ignore
    Progression((SpecificChord.random(), SpecificChord.random()))


def test_init(four_chords):
    a, b, c, d = four_chords
    p = Progression((a, b, c))
    assert len(p) == 3
    assert p[0] == a


def test_all(progression4):
    def check(x, y):
        return x[0] < y[0]
    assert progression4.all([check])
    assert progression4.all_not([lambda x, y: not check(x, y)])


def test_distance(progression4):
    assert progression4.distance == 30


def test_transpose_unique_key(four_chords):
    a, b, c, d = four_chords
    d_ = SpecificChord(frozenset((d.notes_ascending[0] + 12,) + d.notes_ascending[1:]))
    p0 = Progression((a, b, c, d))
    p1 = Progression((a, b, c, d_))
    p2 = Progression(tuple(SpecificChord(frozenset(n + 12 for n in chord.notes)) for chord in p0))
    p3 = Progression(tuple(SpecificChord(frozenset(n + 1 for n in chord.notes)) for chord in p0))
    assert p0.transpose_unique_key() != p1.transpose_unique_key()
    assert p0.transpose_unique_key() == p2.transpose_unique_key()
    assert p0.transpose_unique_key() != p3.transpose_unique_key()
    assert p0.transpose_unique_key(origin_name=False) == p3.transpose_unique_key(origin_name=False)


def test_add_transpose():
    p0 = Progression((
        SpecificChord.from_str('G2_B2_e3'),
        SpecificChord.from_str('A2_C3_E3'),
        SpecificChord.from_str('B2_D3_f3'),
        SpecificChord.from_str('C3_E3_G3'),
    ))
    p1 = Progression((
        SpecificChord.from_str('C0_E0_a0'),
        SpecificChord.from_str('D0_F0_A0'),
        SpecificChord.from_str('E0_G0_B0'),
        SpecificChord.from_str('F0_A0_C1'),
    ))
    assert p0 + -31 == p1
    with pytest.raises(TypeError):
        p0 + [1]  # type: ignore
    assert p0.transposed_to_C0 == p1


def test_sequence(four_chords):
    p = Progression(four_chords)
    assert isinstance(p, Sequence)
    assert p[2] == four_chords[2]
    assert p[1:3] == Progression(four_chords[1:3])



import operator

import pytest

from musiclib import config
from musiclib.chord import Chord
from musiclib.note import Note
from musiclib.scale import ComparedScales
from musiclib.scale import Scale


@pytest.mark.parametrize(
    'notes, name', (
        ('CDEFGAB', 'major'),
        ('CdeFGab', 'phrygian'),
        ('DEFGAbC', 'minor'),
        ('bdefa', 'p_phrygian'),
    ),
)
def test_name(notes, name):
    root = notes[0]
    notes = frozenset(map(Note, notes))
    assert Scale(notes, root=Note(root)).name == name
    assert Scale.from_name(root, name).notes == notes


def test_kind():
    s = Scale.from_name('C', 'major')
    assert (
        s.kind == 'diatonic' and
        {'note_scales', 'triads', 'sevenths', 'ninths', 'notes_to_triad_root'} <= vars(s).keys()
    )
    s = Scale.from_name('C', 'p_major')
    assert (
        s.kind == 'pentatonic' and
        {'note_scales'} <= vars(s).keys() and
        len({'triads', 'sevenths', 'ninths', 'notes_to_triad_root'} & vars(s).keys()) == 0
    )


@pytest.mark.parametrize(
    'op, a, b', (
        (operator.eq, Scale.from_name('C', 'major'), Scale.from_name('C', 'major')),
        (operator.eq, Scale.from_name('C', 'major'), Scale.from_name(Note('C'), 'major')),
        (operator.eq, Scale.from_name(Note('C'), 'major'), Scale.from_name(Note('C'), 'major')),
        (operator.ne, Scale.from_name(Note('C'), 'major'), Scale.from_name(Note('E'), 'major')),
    ),
)
def test_equal(op, a, b):
    assert op(a, b)


@pytest.mark.parametrize(
    'notes, name, nths', (
        ('CDEFGAB', 'triads', 'CEG/C DFA/D EGB/E FAC/F GBD/G ACE/A BDF/B'),
        ('DEFGAbC', 'triads', 'DFA/D EGb/E FAC/F GbD/G ACE/A bDF/b CEG/C'),
        ('CDEFGAB', 'sevenths', 'CEGB/C DFAC/D EGBD/E FACE/F GBDF/G ACEG/A BDFA/B'),
        ('DEFGAbC', 'sevenths', 'DFAC/D EGbD/E FACE/F GbDF/G ACEG/A bDFA/b CEGb/C'),
        ('CDEFGAB', 'ninths', 'CEGBD/C DFACE/D EGBDF/E FACEG/F GBDFA/G ACEGB/A BDFAC/B'),
        ('DEFGAbC', 'ninths', 'DFACE/D EGbDF/E FACEG/F GbDFA/G ACEGb/A bDFAC/b CEGbD/C'),
    ),
)
def test_nths(notes, name, nths):
    assert getattr(Scale.from_str(f'{notes}/{notes[0]}'), name) == tuple(Chord.from_str(s) for s in nths.split())


def test_notes_to_triad_root():
    assert Scale.from_str('DEFGAbC/D').notes_to_triad_root[frozenset(map(Note, 'GbD'))] == 'G'


@pytest.mark.parametrize('notes', ('CDEFGAB', 'BdeEfab', 'deFfabC'))
def test_note_scales(notes):
    assert Scale.from_str(f'{notes}/{notes[0]}').note_scales == dict(zip(notes, config.diatonic, strict=True))


def test_compared():
    left = Scale.from_name('C', 'major')
    right = Scale.from_name('A', 'minor')
    assert ComparedScales(left, right).shared_notes == frozenset(left.notes)

    right = Scale.from_name('C', 'mixolydian')
    c = ComparedScales(left, right)
    assert c.new_notes == frozenset({Note('b')})
    assert c.del_notes == frozenset({Note('B')})


@pytest.mark.parametrize(
    'scale, parallel_name, expected', (
        (Scale.from_name('C', 'major'), 'minor', Scale.from_name('C', 'minor')),
        (Scale.from_name('C', 'minor'), 'major', Scale.from_name('C', 'major')),
        (Scale.from_name('f', 'minor'), 'major', Scale.from_name('f', 'major')),
        (Scale.from_name('C', 'major'), 'phrygian', Scale.from_name('C', 'phrygian')),
        (Scale.from_name('e', 'dorian'), 'locrian', Scale.from_name('e', 'locrian')),
    ),
)
def test_parallel(scale, parallel_name, expected):
    assert scale.parallel(parallel_name) is expected


@pytest.mark.parametrize(
    'scale, relative_name, expected', (
        (Scale.from_name('C', 'major'), 'minor', Scale.from_name('A', 'minor')),
        (Scale.from_name('A', 'minor'), 'major', Scale.from_name('C', 'major')),
        (Scale.from_name('f', 'minor'), 'major', Scale.from_name('A', 'major')),
        (Scale.from_name('A', 'major'), 'minor', Scale.from_name('f', 'minor')),
        (Scale.from_name('C', 'major'), 'phrygian', Scale.from_name('E', 'phrygian')),
        (Scale.from_name('A', 'major'), 'dorian', Scale.from_name('B', 'dorian')),
    ),
)
def test_relative(scale, relative_name, expected):
    assert scale.relative(relative_name) is expected



import dataclasses
import functools

import mido

from musiclib.note import SpecificNote


@dataclasses.dataclass(frozen=True)
@functools.total_ordering
class MidiNote:
    note: SpecificNote
    on: int
    off: int
    track: int

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MidiNote):
            return NotImplemented
        return self.on == other.on

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, MidiNote):
            return NotImplemented
        return self.on < other.on

    def __hash__(self) -> int:
        return hash((self.note, self.track))


def parse_notes(m: mido.MidiFile) -> list[MidiNote]:
    notes: list[MidiNote] = []
    for track_i, track in enumerate(m.tracks):
        t = 0
        t_buffer = {}
        for message in track:
            t += message.time

            if message.type == 'note_on' and message.velocity != 0:
                t_buffer[message.note] = t

            elif message.type == 'note_off' or (
                message.type == 'note_on' and message.velocity == 0
            ):  # https://stackoverflow.com/a/43322203/4204843
                # todo: heapq seems unnecessary here
                # heapq.heappush(
                # notes, MidiNote(
                #     note=SpecificNote.from_i(message.note),
                #     on=t_buffer.pop(message.note),
                #     off=t,
                #     track=track_i,
                # ),
                # )
                notes.append(
                    MidiNote(
                        note=SpecificNote.from_i(message.note),
                        on=t_buffer.pop(message.note),
                        off=t,
                        track=track_i,
                    ),
                )
    return notes


def print_midi(midi: mido.MidiFile) -> None:
    print('n_tracks:', len(midi.tracks))
    print(midi.tracks)
    for i, track in enumerate(midi.tracks):
        print('track', i)
        for message in track:
            print(message)
        print('=' * 100)



import pytest

from musiclib.note import SpecificNote
from musiclib.pitch import Pitch

A3 = SpecificNote('A', 3)
A4 = SpecificNote('A', 4)
A5 = SpecificNote('A', 5)
A6 = SpecificNote('A', 6)
B4 = SpecificNote('B', 4)
b4 = SpecificNote('b', 4)
C4 = SpecificNote('C', 4)
HZ_220 = 220
HZ_440 = 440
HZ_880 = 880
HZ_493 = 493.8833012561241
HZ_452 = 452.8929841231365
HZ_466 = 466.1637615180899
HZ_261 = 261.6255653005986


@pytest.mark.parametrize(
    'i, hz', [
        (-24, 110),
        (-12, HZ_220),
        (-9, HZ_261),
        (0, HZ_440),
        (1, HZ_466),
        (2, HZ_493),
        (12, HZ_880),
    ],
)
def test_i_hz(i, hz):
    pitch = Pitch()
    assert pitch.i_to_hz(i) == hz
    assert int(pitch.hz_to_i(hz)) == i


@pytest.mark.parametrize(
    'note_i, hz', [
        (A3.i, HZ_220),
        (A4.i, HZ_440),
        (b4.i, HZ_466),
        (C4.i, HZ_261),
        (A4.i + 0.5, HZ_452),
    ],
)
def test_note_i_hz(note_i, hz):
    pitch = Pitch()
    assert pitch.note_i_to_hz(note_i) == hz
    assert pitch.hz_to_note_i(hz) == note_i


@pytest.mark.parametrize(
    'hz, note', [
        (441, A4),
        (452.9, b4),
        (465, b4),
    ],
)
def test_round(hz, note):
    pitch = Pitch()
    assert pitch.hz_to_note(hz) == note


@pytest.mark.parametrize(
    'hz_tuning, origin_note, i, hz', [
        (HZ_220, A5, 0, HZ_220),
        (HZ_220, A5, 12, HZ_440),
        (HZ_440, A4, 0, HZ_440),
        (HZ_440, A4, -12, HZ_220),
    ],
)
def test_tuning_origin_note(hz_tuning, origin_note, i, hz):
    pitch = Pitch(hz_tuning, origin_note)
    assert pitch.i_to_hz(i) == hz
    assert int(pitch.hz_to_i(hz)) == i


@pytest.mark.parametrize(
    'note, hz', [
        (A3, HZ_220),
        (A4, HZ_440),
        (A5, HZ_880),
        (C4, HZ_261),
        (B4, HZ_493),
    ],
)
def test_note_hz(note, hz):
    pitch = Pitch()
    assert pitch.note_to_hz(note) == hz
    assert pitch.hz_to_note(hz) == note


@pytest.mark.parametrize(
    'hz, px, hz_min, hz_max, px_max', [
        (55, 0, 55, 880, 1000),
        (110, 250, 55, 880, 1000),
        (220, 500, 55, 880, 1000),
        (440, 750, 55, 880, 1000),
        (880, 1000, 55, 880, 1000),
    ],
)
def test_hz_to_px(hz, px, hz_min, hz_max, px_max):
    assert Pitch.hz_to_px(hz, hz_min, hz_max, px_max) == pytest.approx(px)
    assert Pitch.px_to_hz(px, hz_min, hz_max, px_max) == pytest.approx(hz)


@pytest.mark.parametrize(
    'note, transpose, hz', [
        (A4, -12, HZ_220),
        (A4, +12, HZ_880),
        (A4, +1, HZ_466),
        (A4, +0.5, HZ_452),
    ],
)
def test_transpose(note, transpose, hz):
    pitch = Pitch(transpose=transpose)
    assert pitch.note_to_hz(note) == hz
    assert pitch.hz_to_note(hz) == note


@pytest.mark.parametrize(
    'note, i', [
        (A4, 0),
        (A5, 12),
        (A3, -12),
        (b4, 1),
        (C4, -9),
    ],
)
def test_note_to_i(note, i):
    pitch = Pitch()
    assert pitch.note_to_i(note) == i
    assert pitch.i_to_note(i) == note



import cmath
import math

import svg

from musiclib import config
from musiclib.note import SpecificNote
from musiclib.pitch import Pitch
from musiclib.scale import Scale


class JankoSquare:
    def __init__(
        self,
        scale: Scale = Scale.from_name('C', 'major'),
        origin_note: SpecificNote = SpecificNote('C', 4),
        radius: int = 30,
        n_rows: int = 7,
        n_cols: int = 13,
    ):
        self.width = 2 * n_cols * radius
        self.height = 2 * n_rows * radius
        self.radius = radius
        self.elements: list[svg.Element] = []
        self.scale = scale
        self.origin_note = origin_note
        self.pitch = Pitch(origin_note=origin_note)

        for j in range(-1, n_rows * 2):
            for i in range(-2, n_cols * 2, 2):
                self.add_note(j, i + j % 2)

    def add_note(
        self,
        j: float,
        i: float,
        font_size: int = 15,
    ) -> None:
        note = self.pitch.i_to_note(i)
        x = self.radius + self.radius * i
        y = self.radius + self.radius * j

        if note.abstract in self.scale.note_scales:
            color = config.scale_colors[self.scale.note_scales[note.abstract]]
        else:
            color = config.BLACK_PALE

        polygon = svg.Polygon(
            class_=['polygon-colored'],
            fill=color.css_hex,
            stroke='black',
            stroke_width=1,
            points=self.square_points(x, y, self.radius),  # type: ignore
        )
        self.elements.append(polygon)
        text = svg.Text(
            x=x,
            y=y,
            text=str(note),
            font_size=font_size,
            font_family='monospace',
            text_anchor='middle',
            dominant_baseline='middle',
        )
        text.note = str(note)  # type: ignore
        text.note_midi = note.i  # type: ignore
        self.elements.append(text)

    @staticmethod
    def square_points(x: float, y: float, radius: float) -> list[float]:
        points = []
        for i in range(4):
            theta = math.pi * i / 2
            p = complex(y, x) + radius * cmath.exp(1j * theta)
            points += [p.imag, p.real]
        return points

    def _repr_svg_(self) -> str:
        s = svg.SVG(
            width=self.width,
            height=self.height,
            elements=self.elements,
        )
        return str(s)
